#ifndef WIN32
#include <osmesa.h>
#endif
// begin local includes
/********************************************************************************
*                                                                               *
*                        G L  C o n t e x t E x  C l a s s                      *
*                                                                               *
*********************************************************************************
* Copyright (C) 2003 by ABAQUS, Inc.   All Rights Reserved.                     *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Library General Public                   *
* License as published by the Free Software Foundation; either                  *
* version 2 of the License, or (at your option) any later version.              *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Library General Public License for more details.                              *
*                                                                               *
* You should have received a copy of the GNU Library General Public             *
* License along with this library; if not, write to the Free                    *
* Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.            *
********************************************************************************/
#define FGLDLL_EXPORTS
#include <xincs.h>
#include <fxver.h>
#include <fxdefs.h>
#include <FXStream.h>
#include <FXString.h>
#include <FXObject.h>
#include <FXDict.h>
#include <FXRegistry.h>
#include <FXAccelTable.h>
#include <FXApp.h>
#include <FXId.h>
#include <FXVisual.h>
#include <FXGLVisual.h>
#include <FXOVVisual.h>
#include <FXGLVisualEx.h>
#ifndef WIN32
#include <FXWindow.h>
#include <FXComposite.h>
#include <FXCanvas.h>
#include <FXGLCanvas.h>
#include <FXGLDC.h>
#endif
#include <FXDrawable.h>
#include <FXGLContext.h>
#include <FXGLContextEx.h>
#include <FXGLImage.h>


/*
  Notes:
  - GL Context can be shared between GL canvas widgets.
  - OpenGL documentation says that windows can share GL contexts as long as the
    visuals are the same; so we keep a reference to the GLVisual in the context
    so as to compare it to the visual of the window when the GL context is being 
    locked on the window.
  - Display lists can be shared between different types of GL contexts; the display
    list has to be kept in the same place for each of the contexts that want to share 
    it.  
  - It may be not a bad idea to delay the creation of the context until we call begin().
    This might be much more elegant, as we then already know which surface to use, and
    we won't have to create a temporary dummy window.
  - The mental model of FXGLContext is that of FXDCWindow, except that FXGLContext lives
    long, while FXDCWindow gets created/destroyed on an as-needed basis.
  - There should be NO NEED to derive FXGLContext from FXObject; right now we do, for
    serialization, but I think that we shouldn't; because the stuff in here contains
    info which makes little sense to serialize....
  - Should GL context make sure GLVisual has been create()-ed already?
*/

#define DISPLAY(app) ((Display*)((app)->getDisplay()))

/*******************************************************************************/


// Object implementation
FXIMPLEMENT(FXGLContextEx,FXGLContext,NULL,0)


// For serialization
FXGLContextEx::FXGLContextEx(){
  overlayCtx=0;
  overlayShares=FALSE;
#ifdef WIN32
  imageDC=0;
#endif
  }


// Make a GL context
FXGLContextEx::FXGLContextEx(FXApp* a,FXGLVisual *vis,FXbool wantDirect): // CAE: added wantDirect
  FXGLContext(a,vis,wantDirect),overlayCtx(NULL),overlayShares(FALSE) // CAE: added wantDirect
#ifdef WIN32
      ,imageDC(0)
#endif
  {
  FXTRACE((100,"FXGLContextEx::FXGLContextEx %08x\n",this));
  }


// Make a GL context sharing a display list with another
FXGLContextEx::FXGLContextEx(FXApp* a,FXGLVisual *vis,FXGLContext *shared,FXbool wantDirect): // CAE: added wantDirect
  FXGLContext(a,vis,shared,wantDirect),overlayCtx(NULL),overlayShares(FALSE) // CAE: added wantDirect
#ifdef WIN32
      ,imageDC(0)
#endif
  {
  FXTRACE((100,"FXGLContextEx::FXGLContextEx %08x\n",this));
  }


// Create GL context (using the provided window on Win32)
void FXGLContextEx::create(FXID wnd){
#ifdef HAVE_OPENGL
  if(ctx) return;
#ifndef WIN32
  create();
#else
  if(!wnd){
    create();
    }
  else{
    register FXGLContext *context;
    register void *sharedctx=NULL;
    FXTRACE((100,"FXGLContextEx::create(FXID) %08x\n",this));
  
    // The visual should be OpenGL capable
    if(!visual->info){ fxerror("FXGLContextEx::create(FXID): visual unsuitable for OpenGL.\n"); }

    // Find another member of the group which is already created, and get its context
    if(sgnext!=this){
      context=sgnext;
      while(context!=this){
	sharedctx=context->ctx;
	if(sharedctx) break;
	context=context->sgnext;
	}
      }

    // Get the device context
    HDC hdc = (visual->getFlags() & VISUAL_IMAGE) ? (HDC)wnd : ::GetDC((HWND)wnd);
    
    // Set the pixel format
    if(!SetPixelFormat(hdc,visual->pixelformat,(PIXELFORMATDESCRIPTOR*)visual->info)){
      fxerror("FXGLContextEx::create(): SetPixelFormat(%d) failed.\n", visual->pixelformat); 
      }

    // Make the GL context
    ctx=(void*)wglCreateContext(hdc);
    if(!ctx){ fxerror("FXGLContextEx::create(): wglCreateContext() failed.\n"); }
    xid=(FXID)1;

    // I hope I didn't get this backward; the new context obviously has no
    // display lists yet, but the old one may have, as it has already been around
    // for a while.  If you see this fail and can't explain why, then that might
    // be what's going on.  Report this to jvz@cfdrc.com
    if(sharedctx && !wglShareLists((HGLRC)sharedctx,(HGLRC)ctx)){ fxerror("FXGLContextEx::create(): wglShareLists() failed.\n"); }

    FXGLVisualEx *vis = (FXGLVisualEx *)visual;
    if( 0<vis->getActualOverlaySize() && !overlayCtx ){
      overlayCtx = (void *)wglCreateLayerContext(hdc, 1);
      if(!overlayCtx){ fxerror("FXGLContextEx::create(): wglCreateLayerContext() failed.\n"); }
      // This doesn't always work since it is an RGBA context and a ColorIndex context.
      // Just flag that display lists may not be shared.
      overlayShares = wglShareLists((HGLRC)ctx,(HGLRC)overlayCtx);
      }

    // Release window's device context
    if(!(visual->getFlags() & VISUAL_IMAGE))
      ReleaseDC((HWND)wnd,hdc);
    }
#endif
#endif
  }


// Create GL context
void FXGLContextEx::create(){
  ///visual->create();    // Should we call this here?
#ifdef HAVE_OPENGL
  if(!ctx){
#ifndef WIN32
    register FXGLContext *context;
    register void *sharedctx=NULL;
    FXTRACE((100,"FXGLContextEx::create %08x\n",this));
    
    // The visual should be OpenGL capable
    if(!(VISUAL_OSMESA&visual->getFlags()||visual->info)){ fxerror("FXGLContextEx::create(): visual unsuitable for OpenGL.\n"); }

    // Find another member of the group which is already created, and get its context
    if(sgnext!=this){
      context=sgnext;
      while(context!=this){
        sharedctx=context->ctx;
        if(sharedctx) break;
        context=context->sgnext;
        }
      }

    // Make context
    if(VISUAL_OSMESA&visual->getFlags()){
      ctx=(GLXContext)OSMesaCreateContext(visual->getActualAlphaSize()?OSMESA_RGBA:OSMESA_RGB,
					  (OSMesaContext)sharedctx);
      }
#ifdef GLX_PBUFFER
    else if(VISUAL_PBUFFER&visual->getFlags()){
      ctx=(GLXContext)glXCreateNewContext(DISPLAY(getApp()),*((GLXFBConfig*)visual->info),
					   GLX_RGBA_TYPE, (GLXContext)sharedctx, direct);
      FXTRACE((100,"FXGLContextEx::create PBuffer ctx=%08x\n",ctx));
      }
#endif
    else{
      ctx=glXCreateContext(DISPLAY(getApp()),(XVisualInfo*)visual->info,(GLXContext)sharedctx,
	  VISUAL_IMAGE&visual->getFlags()?FALSE:direct);
      }
    if(!ctx){ fxerror("FXGLContextEx::create(): glXCreateContext() failed.\n"); }
    xid=(FXID)1;

    // Create context for overlay (if requested)
    if(! ((VISUAL_OSMESA|VISUAL_IMAGE)&visual->getFlags()) ){
      FXGLVisualEx *vis = (FXGLVisualEx *)visual;
      if( 0<vis->getActualOverlaySize() && !overlayCtx ){
        FXOVVisual *visOV = vis->getOverlayVisual();
        if(visOV){
          overlayCtx = glXCreateContext(DISPLAY(getApp()),(XVisualInfo*)visOV->info,(GLXContext)ctx,direct);
          if(!overlayCtx){
            // This doesn't always work since it is an RGBA context and a ColorIndex context.
            // Just flag that display lists may not be shared.
            overlayShares = FALSE;
            overlayCtx = glXCreateContext(DISPLAY(getApp()),(XVisualInfo*)visOV->info,NULL,direct);
	    if(!overlayCtx){ fxerror("FXGLContextEx::create(): glXCreateContext() failed for overlay.\n"); }
	    }
	  else{
            overlayShares = TRUE;
	    }
          }
	  FXTRACE((100,"FXGLContextEx::create Overlay context: %x  visual: %x  shares: %s\n", overlayCtx, ((XVisualInfo*)visOV->info)->visualid,overlayShares?"Yes":"No"));
        }
      }
  
#else
  
    // Conceptually, the bitplane organization should be enough to create a context; 
    // but on Windows, there is no concept like a visual, and hence we need to make
    // a dummy window, set its pixel format, and then create the GL context;
    // afterwards, the window is no longer needed and will be deleted.
    // Yes, I'm painfully aware that this sucks, but we intend to keep the 
    // logical model clean come hell or high water....
    HWND wnd=CreateWindow("FXGLCanvas",NULL,WS_CHILD|WS_VISIBLE|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,0,0,2,2,GetDesktopWindow(),0,(HINSTANCE)FXApp::instance()->getDisplay(),this);
    if(!wnd){ fxerror("FXGLContextEx::create(): CreateWindow() failed.\n"); }

    // Jeroen:  I tried to get around the 'clean logical model' vs. 'hell' by
    //          adding a create() overload that would accept a window.  SJG
    create((FXID)wnd);
    
    // Destroy the temporary window
    DestroyWindow(wnd);
#endif
    }
#endif
  }


// Detach the GL context
void FXGLContextEx::detach(){
#ifdef HAVE_OPENGL
  FXGLContext::detach();
  if(overlayCtx){
    FXTRACE((100,"FXGLContextEx::detach %08x\n",this));
    // FIXME Will this leak memory?
    overlayCtx=NULL;
    }
#endif
  }


// Destroy the GL context
void FXGLContextEx::destroy(){
#ifdef HAVE_OPENGL
#ifndef WIN32
  if(VISUAL_OSMESA&visual->getFlags()){
    OSMesaDestroyContext((OSMesaContext)ctx);
    ctx=NULL;
    xid=0;
    }
  else
#endif
    FXGLContext::destroy();
  if(overlayCtx){
#ifndef WIN32
    glXDestroyContext(DISPLAY(getApp()),(GLXContext)overlayCtx);
#else
    wglDeleteContext((HGLRC)overlayCtx);
#endif
    overlayCtx=NULL;
    }
#endif
  }



//  Make the rendering context of drawable current
FXbool FXGLContextEx::beginOverlay(FXDrawable *drawable){
#ifdef HAVE_OPENGL
  if(!drawable){ fxerror("FXGLContextEx::beginOverlay: NULL drawable.\n"); }
  if(visual!=drawable->getVisual()){ fxerror("FXGLContextEx::beginOverlay: visuals do not match.\n"); }
  if(overlayCtx){
    FXTRACE((100,"FXGLContextEx::beginOverlay %08x\n",this));
#ifndef WIN32
    FXGLDC *window = (FXGLDC *)drawable;
    if(!window->getOverlayWindow()){ fxerror("FXGLContextEx::beginOverlay: overlay window not created yet.\n"); }
    if(glXMakeCurrent(DISPLAY(getApp()),window->getOverlayWindow(),(GLXContext)overlayCtx)){
      surface=drawable;
      return TRUE;
      }
#else
    //HDC hdc=drawable->GetDC();  // Obtain DC in a way appropriate for the drawable!
    HDC hdc=::GetDC((HWND)drawable->id());  // FIXME:- it should probably be the line above, but need to check about ReleaseDC first!
    if(wglMakeCurrent(hdc,(HGLRC)overlayCtx)){
      surface=drawable;
      return TRUE;
      }
#endif
    }
#endif
  return FALSE;
  }

void FXGLContextEx::clearOverlay(){
#ifdef HAVE_OPENGL
  if(overlayCtx){
    FXGLVisualEx *vis = (FXGLVisualEx *)visual;
    FXPixel pixel = vis->getOverlayVisual()->getOverlayClearPixel();
    glClearIndex((float)pixel);
    glDrawBuffer(GL_FRONT_AND_BACK);
    glClear(GL_COLOR_BUFFER_BIT);
    }
#endif
  }
  
//  Make the rendering context of drawable non-current
FXbool FXGLContextEx::endOverlay(){
  FXbool bRet=FALSE;
#ifdef HAVE_OPENGL
  if(overlayCtx){
    FXTRACE((100,"FXGLContextEx::endOverlay %08x\n",this));
#ifndef WIN32
    bRet=glXMakeCurrent(DISPLAY(getApp()),None,(GLXContext)NULL);
#else
    if(surface){
      ::ReleaseDC((HWND)surface->id(),wglGetCurrentDC());
      //drawable->ReleaseDC();
      }
    bRet=wglMakeCurrent(NULL,NULL);
#endif
    surface=NULL;
    }
#endif
  return bRet;
  }


// Make OpenGL image context current prior to performing OpenGL commands
FXbool FXGLContextEx::beginImage(FXGLImage *pImage){
#ifdef HAVE_OPENGL
#ifndef WIN32
  if(!pImage){ fxerror("FXGLContextEx::beginImage: NULL image.\n"); }
  if(!pImage->id()){ fxerror("FXGLContextEx::beginImage: image not created yet.\n"); }
  if(visual!=pImage->getVisual()){ fxerror("FXGLContextEx::beginImage: visuals do not match.\n"); }
  create();
  if(ctx){
#ifdef GLX_PBUFFER
    if(VISUAL_PBUFFER&visual->getFlags())
      return glXMakeContextCurrent(DISPLAY(getApp()),pImage->id(),pImage->id(),(GLXContext)ctx);
    else
#endif
    if(! (VISUAL_OSMESA&visual->getFlags()) )
      return glXMakeCurrent(DISPLAY(getApp()),pImage->id(),(GLXContext)ctx);
    else
      return OSMesaMakeCurrent((OSMesaContext)ctx,pImage->getData(),GL_UNSIGNED_BYTE,
          pImage->getWidth(),pImage->getHeight());
    }
  return FALSE;
#else
  imageDC = pImage->GetDC();
  if(imageDC){
    create(imageDC);
    // Set the pixel format
    if(!SetPixelFormat((HDC)imageDC,visual->pixelformat,(PIXELFORMATDESCRIPTOR*)visual->info)){ 
      fxerror("FXGLContextEx::beginImage(): SetPixelFormat() failed.\n"); 
      }
    if(visual->hPalette){
      SelectPalette((HDC)imageDC,(HPALETTE)visual->hPalette,FALSE);
      RealizePalette((HDC)imageDC);
      }
    return wglMakeCurrent((HDC)imageDC,(HGLRC)ctx);
    }
#endif
#endif
  return false;
  }


// Make OpenGL image context non current 
FXbool FXGLContextEx::endImage(FXGLImage *pImage){
  FXbool bSts = true;
#ifndef WIN32
  if(! (VISUAL_OSMESA&visual->getFlags()) )
#endif
    bSts = end();
#ifdef HAVE_OPENGL
#ifdef WIN32
  if(0 == pImage->ReleaseDC(imageDC))
    bSts = FALSE;
#endif
#endif
  return bSts;
  }



// Close and release any resources
FXGLContextEx::~FXGLContextEx(){
  FXTRACE((100,"FXGLContextEx::~FXGLContextEx %08x\n",this));
#ifdef HAVE_OPENGL
  if(overlayCtx){
#ifndef WIN32
    glXDestroyContext(DISPLAY(getApp()),(GLXContext)overlayCtx);
#else
    wglDeleteContext((HGLRC)overlayCtx);
#endif
    overlayCtx=(FXGLContext*)-1;
    }
#ifndef WIN32
  if(ctx && (VISUAL_OSMESA&visual->getFlags())){
    OSMesaDestroyContext((OSMesaContext)ctx);
    ctx=NULL; // So parent destructor will not crash
    }
#endif
#endif
  }
