// -*- Mode: C++ -*-
#ifndef FXTheme_h
#define FXTheme_h
// Begin local includes
#include <fxdefs.h>
#include <FXSize.h>

class FXDCWindow;
class FXWindow;
class FXString;

enum FXThemeClass {
    THEME_NONE,
    THEME_ANY,
    THEME_BUTTON,
    THEME_COMBOBOX,
    THEME_GROUPBOX,
    THEME_EDIT,
    THEME_HEADER,
    THEME_PROGRESS,
    THEME_SCROLLBAR,
    THEME_SLIDER,
    THEME_SPINNER,
    THEME_TAB,
    THEME_TOOLBAR,
    THEME_TREE,
    THEME_TOPWINDOW,
    THEME_LAST,
    THEME_INACTIVE
};

// This class is essentially an interface to the Windows XP theming engine.
// All calls are ignored on X and on Windows with the Classic look.
class FXAPI FXTheme
{
public:

    enum Orientation {
        UP,
        DOWN,
        LEFT,
        RIGHT
    };

    // Initializes themes.
    static void initialize();

    // Finalizes themes.
    static void finalize();

    // Constructs a theme object for the given window and theme class.
    FXTheme( FXWindow *window, FXuint cls );

    // Constructs a theme object for the given window. The theme object
    // must be reset() with appropriate class name before using.
    FXTheme( FXWindow *window );

    // Destructor.
    ~FXTheme();

    // Returns TRUE if this theme object can be used for drawing.
    FXbool active() const { return m_theme != NULL; }

    // Resets the theme to the given class.
    void reset( FXuint cls = THEME_ANY );

    // Destroys resources associated with this object.
    void destroy();

    // Draws the given part. Optionally, may draw the part rotated.
    // See fxtmschema.h for available parts and states.
    void drawBackground(
        FXDCWindow &dc,
        FXint partId,
        FXint stateId,
        FXint x = 0,
        FXint y = 0,
        FXint w = 0,
        FXint h = 0,
        FXint clipX = 0,
        FXint clipY = 0,
        FXint clipW = 0,
        FXint clipH = 0,
        Orientation orientation = UP
        );

    // Draws an edge (frame). Frame style is given by FRAME_xxx constants.
    void drawEdge(
        FXDCWindow &dc,
        int partId,
        int stateId,
        FXuint frameStyle,
        int x = 0,
        int y = 0,
        int w = 0,
        int h = 0
        );

    // Set the shape of the part to be the shape of the window.
    // This function may be applied to non-rectangular windows
    // that must overlap with siblings, such as MDI children.
    void setWindowRegion(
        FXDCWindow &dc,
        int partId,
        int stateId,
        int x = 0,
        int y = 0,
        int w = 0,
        int h = 0
        );

    // Clears the window region set previously with setWindowRegion.
    void clearWindowRegion();

    // Add the region occupied by the given part to the clipping
    // region of the DC. Subsequent drawing operations will leave
    // the excluded region untouched.
    void excludePartRegion(
        FXDCWindow &dc,
        int partId,
        int stateId,
        int x = 0,
        int y = 0,
        int w = 0,
        int h = 0
        );

    // Returns the default size of the given part.
    FXSize getDefaultSize( int partId, int stateId );

    // Returns the text color of the given part.
    // On XP used for group boxes.
    FXColor getThemeTextColor( int partId, int stateId );

    // Returns the menu background color.
    FXColor getSysMenuColor();

    // Returns the background color for highlighted menu items.
    FXColor getSysMenuHighlight();

    // Returns the color of the top window title text.
    FXColor getSysCaptionTextColor();

    // Returns the color of the inactive top window title text.
    FXColor getSysInactiveCaptionTextColor();

    // Returns the color of the MDI client.
    FXColor getSysAppWorkspaceColor();

    // Returns the width of the vertical scrollbars.
    FXint getSysVerticalScrollbarWidth();

    // Returns the height of horizontal scrollbars.
    FXint getSysHorizontalScrollbarHeight();

    // Offsets the origin of the given DC by the given amounts.
    // This function is not available in the FXDC class, but is essential
    // for the simulated transparency used in XP themes.
    void offsetOrigin( FXDCWindow &dc, int dx, int dy );

private:
    FXWindow *m_window;
    FXID m_theme;
    FXuint m_class;
};


#endif
