// begin local includes
/********************************************************************************
*                                                                               *
*                           S t r i n g   O b j e c t                           *
*                                                                               *
*********************************************************************************
* Copyright (C) 1997,2002 by Jeroen van der Zijp.   All Rights Reserved.        *
*********************************************************************************
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU             *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.    *
*********************************************************************************
* $Id: FXString.cpp,v 1.64.4.3 2003/01/09 18:35:37 fox Exp $                        *
********************************************************************************/
#ifdef HAVE_VSSCANF
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#endif
#include <xincs.h>
#include <fxver.h>
#include <fxdefs.h>
#include <FXStream.h>
#include <FXString.h>


/*
  Notes:
  - The special pointer-value null represents an empty "" string.
  - A call to strlen(str) is optimized to i386 scansb instruction; we call strlen() internally,
    rather than using length().
  - Strings are never NULL:- this speeds things up a lot.
  - Whenever an empty string "" would result, we try to substitute that with null,
    rather than keep a buffer.
*/



// The string buffer is always rounded to a multiple of ROUNDVAL
// which must be 2^n.  Thus, small size changes will not result in any
// actual resizing of the buffer except when ROUNDVAL is exceeded.
#define ROUNDVAL    16

// Round up to nearest ROUNDVAL
#define ROUNDUP(n)  (((n)+ROUNDVAL-1)&-ROUNDVAL)

// Special NULL string
const FXchar FXString::null[4]={0,0,0,0};

// Numbers for hexadecimal
const FXchar FXString::hex[17]="0123456789abcdef";
const FXchar FXString::HEX[17]="0123456789ABCDEF";


// Size to capacity
void FXString::size(FXint sz){
	// never resize smaller
	if (sz < Length())
		sz = Length() +1;
    isolate(sz);
}


// Construct and fill with constant
FXString::FXString(FXchar c,FXint n)
:atr_StringBase()
{
  if(0<n){
    isolate(n);
    char *buff = GetBuffPtr();
    memset(buff,c,n);
    buff[n]=ATR_EOS;
    SetStringLength(n);
    }
}


// Construct string from two parts
FXString::FXString(const FXchar *s1,const FXchar* s2){
    size_t slen1 = 0;
    size_t slen2 = 0;
    if (s1 && s1[0])
		slen1 = strlen(s1);
    if (s2 && s2[0])
		slen2 = strlen(s2);
    isolate(slen1+slen2 + 1);
    if (slen1) 
        append(s1);
    if (slen2) 
        append(s2);
  }


// Extract partition of delimiter separated string
FXString FXString::extract(FXint part,FXchar delim) const {
  FXString res;
  if (length()) {
      register const FXchar *s=CStr();
      while(s[0] && part){part-=(s[0]==delim);s++;}
      if(s[0]){
	register FXint len=0;
	while(s[len] && s[len]!=delim) len++;
	if(0<len){
	  res.isolate(len+1);
	  char * buff = res.GetBuffPtr();
	  memcpy(buff,s,len);
	  buff[len]=ATR_EOS;
	  res.SetStringLength(len);
	}
      }
  }
  return res;
}


// Extract partition of string, interpreting escapes
FXString FXString::extract(FXint part,FXchar delim,FXchar esc) const {
  FXString res;
  if (length()) {
      register const FXchar *s=CStr();
      while(s[0] && part){part-=(s[0]==delim);s++;}
      if(s[0]){
	register FXint len=0;
	while(s[len] && s[len]!=delim) len++;
	if(0<len){
	  register FXint i,j;
	  res.isolate(len+1);
	  char * buff = res.GetBuffPtr();
	  for(i=j=0; j<len; j++){
	    if(s[j]==esc){ if(s[j+1]!=esc) continue; j++; }
	    buff[i++]=s[j];
	  }
	  buff[i]=ATR_EOS;
	  res.SetStringLength(i);
	}
      }
  }
  return res;
  }

FXint 
FXString::size() const
{
    if (m_StringRep == 0)
	return 0;
    int len = strlen(CStr());
    if (len)
	return len + 1;

    return len;
}

FXchar& FXString::operator [](FXint ix) 
{
  //    FXASSERT(ix <= length());
    isolate(Length());
    return GetBuffPtr()[ix];
}



// Concatenate two FXStrings
FXString operator+(const FXString& s1,const FXString& s2){
  return FXString(s1.CStr(),s2.CStr());
  }


// Concatenate FXString and string
FXString operator+(const FXString& s1,const FXchar* s2){
  return FXString(s1.CStr(),s2);
  }


// Concatenate string and FXString
FXString operator+(const FXchar* s1,const FXString& s2){
  return FXString(s1,s2.CStr());
  }


// Concatenate FXString and character
FXString operator+(const FXString& s,FXchar c){
    FXString res;
    size_t len = s.length();
    res.isolate(len + 2);
    char * buff = res.GetBuffPtr();
    if (len) 
	memcpy(buff,s.text(),len);
    buff[len] = c;
    buff[len+1] = ATR_EOS;
    res.SetStringLength(len+1);
    return res;
  }


// Concatenate character and FXString
FXString operator+(FXchar c,const FXString& s){
    FXString res;
    size_t len = s.length();
    res.isolate(len + 2);
    char * buff = res.GetBuffPtr();
    buff[0] = c;
    if (len)
	memcpy(buff+1,s.text(),len);
    buff[len+1] = ATR_EOS;
    res.SetStringLength(len+1);
    return res;
  }


// Fill with a constant
FXString& FXString::fill(FXchar c,FXint n){
  if(0<n){
     delete get( n+1);
     char * buff = GetBuffPtr();
     memset(buff,c,n);
      buff[n]=ATR_EOS;
      SetStringLength(n);
  }
  return *this;
  }


// Fill up to current length
FXString& FXString::fill(FXchar c){
  int n(length());
  if(n){
    delete get(n+1);
    char * buff = GetBuffPtr();
    memset(buff,c,n);
    buff[n]=ATR_EOS;
    SetStringLength(n);
    }
  return *this;
  }


// Insert character at position
FXString& FXString::insert(FXint pos,FXchar c){
  register FXint len(length());
  if(pos<=0)
    atr_StringBase::insert(0,&c,1,0,1);
  else if(pos>=len)
      atr_StringBase::append(&c,1,0,1);
  else
      atr_StringBase::insert(pos,&c,1,0,1);
  return *this;
  }


// Insert n characters c at specified position
FXString& FXString::insert(FXint pos,FXchar c,FXint n){
  if(0<n){
	register FXint len(length());
	if (pos>=len) {
	    isolate(len+n+1);
	    char * buff = GetBuffPtr();
	    memset(&buff[len],c,n);
	    buff[len+n]=ATR_EOS;		
	}
	else if(pos<=0){ 
		pos = 0;
	    char * oldBuff = GetBuffPtr();
	    char* trash = get(len+n+1);
	    char * newBuff = GetBuffPtr();
	    if (newBuff == oldBuff) {
	      memmove(&newBuff[n],&newBuff[0],len+1);
	      memset(&newBuff[pos],c,n);
		}
	    else {
	      memcpy(&newBuff[n],&oldBuff[0],len+1);
	      memset(&newBuff[0],c,n);
	      delete [] trash;
	    }
	}
	else {
	    isolate(len+n+1);
	    char * buff = GetBuffPtr();
	    memmove(&buff[pos+n],&buff[pos],len-pos+1);
	    memset(&buff[pos],c,n);
	}
	SetStringLength(len+n);
  }
  return *this;
  }


// Insert string at position
FXString& FXString::insert(FXint pos,const FXchar* s,FXint n){
  if(0<n && s){
	  register FXint len(length());
	  if(pos<=0) 
		atr_StringBase::insert(0,s, n,0,n);
	  else if(pos>=len)
		  atr_StringBase::append(s,n,0,n);
	  else
		  atr_StringBase::insert(pos,s,n,0,n);
	  if (pos>=len) {
	    char * buff = GetBuffPtr();
	    buff[len+n]=ATR_EOS;		
	    SetStringLength(len+n);
	  }
  }
  return *this;
}
	  

// Insert string at position
 FXString& FXString::insert(FXint pos,const FXString& s){
  register FXint len(s.length());
  if (len)
      insert(pos,s.CStr(),len);
  return *this;
  }


// Insert string at position
 FXString& FXString::insert(FXint pos,const FXchar* s){
  if (s)
      insert(pos,s,strlen(s));
  return *this;
  }



// Append input n characters c to this string
FXString& FXString::append(FXchar c,FXint n){
  if(0<n){
	register FXint len(length());
    isolate(len+n+1);
    char *buff = GetBuffPtr();
    memset(&buff[len],c,n);
    buff[len+n]=ATR_EOS;
    SetStringLength(len+n);
    }
  return *this;
  }


// Add string to the end
FXString& FXString::append(const FXchar *s,FXint n){
  if(0<n && s)
 	  atr_StringBase::append(s,n,0,n);   
  return *this;
  }


// Add string to the end
FXString& FXString::append(const FXString& s){
  register FXint len(s.length());
  if (len)
      append(s.CStr(),len);
  return *this;
  }


// Add string to the end
FXString& FXString::append(const FXchar *s){
  if (s)
      append(s,strlen(s));
  return *this;
  }



// Append string
FXString& FXString::operator+=(const FXchar* s){
  if (s)
      append(s,strlen(s));
  return *this;
  }



// Prepend string
FXString& FXString::prepend(const FXchar *s,FXint n){
  if(0<n && s)
	insert(0,s,n);
  return *this;
  }


// Prepend string
FXString& FXString::prepend(const FXchar *s){
  if (s)
      insert(0,s);
  return *this;
  }

FXString& FXString::replace(FXint pos,FXchar c){
  if (pos < 0)
    prepend(c);
  else if (pos > length())
    append(c);
  else
      atr_StringBase::replace(pos,1,&c,1,0, atr_StrEnd);
  return *this;
  }


// Replace character in string

// Replace part of string
FXString& FXString::replace(FXint pos,FXint m,const FXchar *s,FXint n){
  int len = length();
  if(0<= n){
	if (pos>=len || m==0) {
	    insert(pos,s,n);
	}
	else if(pos+m<=0){ 
	    isolate(len+1);
	    char * buff = GetBuffPtr();
	    memcpy(buff,s,n);
	    SetStringLength(len);
	}
	else {
	    if(pos<0){m+=pos;pos=0;}
	    if(pos+m>len){m=len-pos;}
	    isolate(len+n+1);
	    char * buff = GetBuffPtr();
	    memmove(&buff[pos+n],&buff[pos+m],len-pos-m+1);
	    memcpy(&buff[pos],s,n);
	    SetStringLength(len-m+n);
	}
  }
  return *this;
  }


// Replace the m characters at pos with n characters c
FXString& FXString::replace(FXint pos,FXint m,FXchar c,FXint n){
  int len = length();
  if(0<=n){
	if (pos>=len || m==0) {
	    insert(pos,c,n);
	}
	else if(pos+m<=0){ 
	    isolate(len+1);
	    char * buff = GetBuffPtr();
	    memset(buff,c,n);
	    SetStringLength(len);
	}
	else {
	    if(pos<0){m+=pos;pos=0;}
	    if(pos+m>len){m=len-pos;}
	    isolate(len+n+1);
	    char * buff = GetBuffPtr();
	    memmove(&buff[pos+n],&buff[pos+m],len-pos-m+1);
	    memset(&buff[pos],c,n);
	    SetStringLength(len-m+n);
	}
  }
  return *this;
  }

// Replace part of string
FXString& FXString::replace(FXint pos,FXint m,const FXchar *s){
  if (s)
      replace(pos,m,s,strlen(s));
  return *this;
  }



// Remove section from buffer
FXString& FXString::remove(FXint pos,FXint n){
    if(0<n){
      register FXint len = length();
      if(pos<len && pos+n>0){
	if(pos<0){n+=pos;pos=0;}
	if(pos+n>len){n=len-pos;}
	if (0 < n)
		atr_StringBase::Erase(pos,n);
      }
    }
    return *this;
}
    


// Substitute one character by another
FXString& FXString::substitute(FXchar orig,FXchar sub){
  register FXint len(length());
  if(len){
    isolate(len +1);
    register FXchar *s=GetBuffPtr();
    do{ if(*s==orig){ *s=sub; } }while(*++s);
    SetStringLength(len);
    }
  return *this;
  }


// Remove leading and trailing whitespace
FXString& FXString::trim(){
  register FXint len(length());
  if(len){
    isolate(len +1);
    register FXint s=0;
    register FXint e=len;
    char *buff = GetBuffPtr();
    while(0<e && isspace((FXuchar)buff[e-1])) e--;
    while(s<e && isspace((FXuchar)buff[s])) s++;
    if(s<e) memmove(buff,&buff[s],e-s);
    buff[e-s]=ATR_EOS;
    SetStringLength(e-s);
    }
  return *this;
  }


// Remove leading whitespace
FXString& FXString::trimBegin(){
  register FXint len(length());
  if(len){
    isolate(len +1);
    register FXint s=0;
    register FXint x=0;
	char *buff = GetBuffPtr();
    while(isspace((FXuchar)buff[s])) s++;
    while(buff[s]) buff[x++]=buff[s++];
    buff[x]=ATR_EOS;
    SetStringLength(x);
    }
  return *this;
  }


// Remove trailing whitespace
FXString& FXString::trimEnd(){
  register FXint len(length());
  if(len){
    isolate(len +1);
    register FXint e=len;
	char *buff = GetBuffPtr();
    while(0<e && isspace((FXuchar)buff[e-1])) e--;
    buff[e]=ATR_EOS;
    SetStringLength(e);
    }
  return *this;
  }


// Truncate string
FXString& FXString::trunc(FXint pos){
  register FXint len(length());
  if(pos<0) pos=0;
  if(pos<len) {
    atr_StringBase::Erase(pos,atr_StrEnd );
  }
  return *this;
  }



// Get leftmost part
FXString FXString::left(FXint n) const {
  FXString res;
  if(0<n){
    register FXint len=length();
    if(n>len) n=len;
    return FXString(CStr(),n);
    }
  return res;
  }


// Get rightmost part
FXString FXString::right(FXint n) const {
  FXString res;
  if(0<n){
    register FXint len=length();
    if(n>len) n=len;
	const char* buff=CStr();
    return FXString(&buff[len-n],n);
    }
  return res;
  }


// Get some part in the middle
FXString FXString::mid(FXint pos,FXint n) const {
  FXString res;
  if(0<n){
    register FXint len=length();
    if(pos<len && pos+n>0){
      if(pos<0){n+=pos;pos=0;}
      if(pos+n>len){n=len-pos;}
	  const char* buff = CStr();
      return FXString(&buff[pos],n);
      }
    }
  return res;
  }

// Return all characters before the nth occurrence of ch,
// counting from the beginning of the string if n>0, or
// from the end if n<0.
FXString FXString::before(FXchar c,FXint n) const {
  FXString res;
  register const FXchar *end;
  register FXchar cc=c;
  register FXint len=length();
  const char * buff = CStr();
  if(n && len>0){
    if(0<n){
      end=CStr();
      while(*end){
        if(*end==cc){
          if(--n==0) break;
          }
        end++;
        }
      }
    else{
      end=buff + len;
      while(buff<end){
        end--;
        if(*end==cc){
          if(++n==0) break;
          }
        }
      }
    return FXString(buff,end-buff);
    }
  return res;
  }



// Return all characters after the nth occurrence of ch,
// counting from the beginning of the string if n>0, or
// from the end if n<0.
FXString FXString::after(FXchar c,FXint n) const {
  FXString res;
  register const FXchar *beg,*end;
  register FXchar cc=c;
  register FXint len=length();
  if(n && len){
    const char * buff = CStr();
    end=buff + len;
    if(0<n){
      beg=buff;
      while(*beg){
        beg++;
        if(*(beg-1)==cc){
          if(--n==0) break;
          }
        }
      }
    else{
      beg=end;
      while(buff<beg){
        if(*(beg-1)==cc){
          if(++n==0) break;
          }
        beg--;
        }
      }
    return FXString(beg,end-beg);
    }
  return res;
  }

// Compare strings
FXint compare(const FXchar *s1,const FXchar *s2){
  register const FXuchar *p1=(const FXuchar *)s1;
  register const FXuchar *p2=(const FXuchar *)s2;
  register FXint c1,c2;
  do{
    c1=*p1++;
    c2=*p2++;
    }
  while(c1 && (c1==c2));
  return c1-c2;
  }

FXint compare(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text());
  }

FXint compare(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2);
  }

FXint compare(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text());
  }


// compare strings up to n
FXint compare(const FXchar *s1,const FXchar *s2,FXint n){
  register const FXuchar *p1=(const FXuchar *)s1;
  register const FXuchar *p2=(const FXuchar *)s2;
  register FXint c1,c2;
  if(0<n){
    do{
      c1=*p1++;
      c2=*p2++;
      }
    while(--n && c1 && (c1==c2));
    return c1-c2;
    }
  return 0;
  }

FXint compare(const FXchar *s1,const FXString &s2,FXint n){
  return compare(s1,s2.text(),n);
  }

FXint compare(const FXString &s1,const FXchar *s2,FXint n){
  return compare(s1.text(),s2,n);
  }

FXint compare(const FXString &s1,const FXString &s2,FXint n){
  return compare(s1.text(),s2.text(),n);
  }


// Compare strings case insensitive
FXint comparecase(const FXchar *s1,const FXchar *s2){
  register const FXuchar *p1=(const FXuchar *)s1;
  register const FXuchar *p2=(const FXuchar *)s2;
  register FXint c1,c2;
  do{
    c1=tolower(*p1++);
    c2=tolower(*p2++);
    }
  while(c1 && (c1==c2));
  return c1-c2;
  }

FXint comparecase(const FXchar *s1,const FXString &s2){
  return comparecase(s1,s2.text());
  }

FXint comparecase(const FXString &s1,const FXchar *s2){
  return comparecase(s1.text(),s2);
  }

FXint comparecase(const FXString &s1,const FXString &s2){
  return comparecase(s1.text(),s2.text());
  }


// Compare strings case insensitive up to n
FXint comparecase(const FXchar *s1,const FXchar *s2,FXint n){
  register const FXuchar *p1=(const FXuchar *)s1;
  register const FXuchar *p2=(const FXuchar *)s2;
  register FXint c1,c2;
  if(0<n){
    do{
      c1=tolower(*p1++);
      c2=tolower(*p2++);
      }
    while(--n && c1 && (c1==c2));
    return c1-c2;
    }
  return 0;
  }

FXint comparecase(const FXchar *s1,const FXString &s2,FXint n){
  return comparecase(s1,s2.text(),n);
  }

FXint comparecase(const FXString &s1,const FXchar *s2,FXint n){
  return comparecase(s1.text(),s2,n);
  }

FXint comparecase(const FXString &s1,const FXString &s2,FXint n){
  return comparecase(s1.text(),s2.text(),n);
  }

/// Comparison operators
FXbool operator==(const FXString &s1,const FXString &s2){
  return compare(s1.CStr(),s2.CStr())==0;
  }

FXbool operator==(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)==0;
  }

FXbool operator==(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())==0;
  }

FXbool operator!=(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text())!=0;
  }

FXbool operator!=(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)!=0;
  }

FXbool operator!=(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())!=0;
  }

FXbool operator<(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text())<0;
  }

FXbool operator<(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)<0;
  }

FXbool operator<(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())<0;
  }

FXbool operator<=(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text())<=0;
  }

FXbool operator<=(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)<=0;
  }

FXbool operator<=(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())<=0;
  }

FXbool operator>(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text())>0;
  }

FXbool operator>(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)>0;
  }

FXbool operator>(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())>0;
  }

FXbool operator>=(const FXString &s1,const FXString &s2){
  return compare(s1.text(),s2.text())>=0;
  }

FXbool operator>=(const FXString &s1,const FXchar *s2){
  return compare(s1.text(),s2)>=0;
  }

FXbool operator>=(const FXchar *s1,const FXString &s2){
  return compare(s1,s2.text())>=0;
  }


// Find a character, searching forward; return position or -1
FXint FXString::findf(FXchar c,FXint pos) const {
  if(0<=pos){
    register FXint len=length();
    register FXint p=pos;
    register FXchar cc=c;
    register const FXchar *buff=CStr();
    while(p<len){ if(buff[p]==cc){ return p; } p++; }
    }
  return -1;
  }


// Find a character, searching backward; return position or -1
FXint FXString::findb(FXchar c,FXint pos) const {
  if(0<=pos){
    register FXint len=length();
    register FXint p=pos;
    register FXchar cc=c;
    if(p>=len) p=len-1;
    register const FXchar *buff=CStr();
    while(0<=p){ if(buff[p]==cc){ return p; } p--; }
    }
  return -1;
  }


// Find a substring of length n, searching forward; return position or -1
FXint FXString::findf(const FXchar* substr,FXint n,FXint pos) const {
  register FXint len=length();
  if(0<=pos && 0<n && n<=len){
    register FXchar c=substr[0];
    len=len-n+1;
    register const FXchar *buff=CStr();
    while(pos<len){
      if(buff[pos]==c){
        if(!compare(buff+pos,substr,n)){
          return pos;
          }
        }
      pos++;
      }
    }
  return -1;
  }


// Find a substring of length n, searching backward; return position or -1
FXint FXString::findb(const FXchar* substr,FXint n,FXint pos) const {
  register FXint len=length();
  if(0<=pos && 0<n && n<=len){
    register FXchar c=substr[0];
    len-=n;
    if(pos>len) pos=len;
    register const FXchar *buff=CStr();
    while(0<=pos){
      if(buff[pos]==c){
        if(!compare(buff+pos,substr,n)){
          return pos;
          }
        }
      pos--;
      }
    }
  return -1;
  }

// Find a substring, searching forward; return position or -1
FXint FXString::findf(const FXchar* substr,FXint pos) const {
  return findf(substr,strlen(substr),pos);
  }

// Find a substring, searching backward; return position or -1
FXint FXString::findb(const FXchar* substr,FXint pos) const {
  return findb(substr,strlen(substr),pos);
  }

// Find a substring, searching forward; return position or -1
FXint FXString::findf(const FXString &substr,FXint pos) const {
  return findf(substr.text(),substr.length(),pos);
  }

// Find a substring, searching backward; return position or -1
FXint FXString::findb(const FXString &substr,FXint pos) const {
  return findb(substr.text(),substr.length(),pos);
  }


// Find number of occurances of character in string
FXint FXString::count(FXchar c) const {
  register FXint len=length();
  register FXint n=0;
  if(len){
	  register const FXchar *s=CStr();
	  while(*s){ n+=(c==*s); s++; }
  }
  return n;
  }


// Get hash value
FXint FXString::hash() const {
  register FXint len=length();
  register FXint h=0;
  if(len){
	  register const FXchar *s=CStr();
	  register FXint g;
	  while(*s) {
		h=(h<<4)+*s++;
		g=h&0xF0000000UL;
		if(g) h^=g>>24;
		h&=~g;
		}
	}
  return h;
  }


// Save
FXStream& operator<<(FXStream& store,const FXString& s){
  FXint sz=s.size();
  store << sz;
  store << s.length();
  if(sz){store.save(s.CStr(),sz);}
  return store;
  }


// Load
FXStream& operator>>(FXStream& store,FXString& s){
  FXint sz;
  store >> sz;
  s.size(sz);
  if(sz){
      FXint len;
      store >> len;
      char * buff = s.GetBuffPtr();
      store.load(buff,len +1);
      buff[len] = 0;
      s.SetStringLength(len);
  }
  return store;
  }


// Print formatted string a-la vprintf
FXString& FXString::vformat(const char *fmt,va_list args){
  register FXint len=0;
  if(fmt && *fmt){
    register FXint s=strlen(fmt);       // Result is longer than format string
#if (__GLIBC__ >= 2) || defined(WIN32)
    s+=128;
x:  size(s);
    len=vsnprintf(GetBuffPtr(),s,fmt,args);
    if(len<0){ s<<=1; goto x; }         // Some implementations return -1 if not enough room
    if(s<=len){ s=len+1; goto x; }      // Others return how much space would be needed
#else
    s+=1024;                            // Add a lot of slop
    size(s);
    len=vsprintf(GetBuffPtr(),fmt,args);
#endif
    }
  size(len+1);
  SetStringLength(len);
  return *this;
  }


// Print formatted string a-la printf
FXString& FXString::format(const char *fmt,...){
  va_list args;
  va_start(args,fmt);
  vformat(fmt,args);
  va_end(args);
  return *this;
  }


// Furnish our own version if we have to
#ifndef HAVE_VSSCANF
extern "C" int vsscanf(const char* str, const char* format, va_list arg_ptr);
#endif


// Scan
FXint FXString::vscan(const char *fmt,va_list args) const {
  register FXint len=length();
  if(len)
	  return vsscanf((char*)CStr(),fmt,args); // Cast needed for HP-UX 11, which has wrong prototype for vsscanf
  return 0;
  }


FXint FXString::scan(const char *fmt,...) const {
  FXint result;
  va_list args;
  va_start(args,fmt);
  result=vscan(fmt,args);
  va_end(args);
  return result;
  }


// Format a string a-la vprintf
FXString FXStringVFormat(const FXchar *fmt,va_list args){
  FXString result;
  result.vformat(fmt,args);
  return result;
  }


// Format a string a-la printf
FXString FXStringFormat(const FXchar *fmt,...){
  FXString result;
  va_list args;
  va_start(args,fmt);
  result.vformat(fmt,args);
  va_end(args);
  return result;
  }


// Conversion of integer to string
FXString FXStringVal(FXint num,FXint base){
  FXchar buf[34];
  register FXchar *p=buf+33;
  register FXuint nn=(FXuint)num;
  if(base<2 || base>16){fxerror("FXStringVal: base out of range\n");}
  if(num<0){nn=(FXuint)(~num)+1;}
  *p='\0';
  do{
    *--p=FXString::HEX[nn%base];
    nn/=base;
    }
  while(nn);
  if(num<0) *--p='-';
  FXASSERT(buf<=p);
  return FXString(p,buf+33-p);
  }


// Conversion of unsigned integer to string
FXString FXStringVal(FXuint num,FXint base){
  FXchar buf[34];
  register FXchar *p=buf+33;
  register FXuint nn=num;
  if(base<2 || base>16){fxerror("FXStringVal: base out of range\n");}
  *p='\0';
  do{
    *--p=FXString::HEX[nn%base];
    nn/=base;
    }
  while(nn);
  FXASSERT(buf<=p);
  return FXString(p,buf+33-p);
  }


// Formatting for reals
static const char *const expo[]={"%.*f","%.*E","%.*G"};


// Conversion of float to string
FXString FXStringVal(FXfloat num,FXint prec,FXbool exp){
  return FXStringFormat(expo[exp],prec,num);
  }


// Conversion of double to string
FXString FXStringVal(FXdouble num,FXint prec,FXbool exp){
  return FXStringFormat(expo[exp],prec,num);
  }


// Value of digit 0-9,A-F
static inline FXint digitvalue(FXchar c){
  return (c<='9') ? (FXint)(c-'0') : (FXint)(toupper(c)-'A'+10);
  }


// Conversion of string to integer
FXint FXIntVal(const FXString& s,FXint base){
  return (FXint)strtol(s.CStr(),NULL,base);
  }


// Conversion of string to unsigned integer
FXuint FXUIntVal(const FXString& s,FXint base){
  return (FXuint)strtoul(s.CStr(),NULL,base);
  }


// Conversion of string to float
FXfloat FXFloatVal(const FXString& s){
  return (FXfloat)strtod(s.CStr(),NULL);
  }


// Conversion of string to double
FXdouble FXDoubleVal(const FXString& s){
  return strtod(s.CStr(),NULL);
  }



